import unittest

import pandas as pd
from racing_telemetry import Telemetry
from racing_telemetry.retrieval.retrieval_strategy import RetrievalStrategy
from racing_telemetry.retrieval.graphql_retrieval_strategy import GraphQLRetrievalStrategy

class MockDataRetrievalStrategy(RetrievalStrategy):
    def retrieve_data(self, filters):
        return [{"key": "value"}]

class TestTelemetry(unittest.TestCase):

    # def test_get_data(self):
    #     strategy = MockDataRetrievalStrategy()
    #     telemetry = Telemetry(strategy)
    #     data = telemetry.get_data()
    #     self.assertEqual(data, [{"key": "value"}])

    # def test_graphql_retrieval_strategy(self):
    #     strategy = GraphQLRetrievalStrategy(endpoint="http://telemetry.b4mad.racing:30050/graphql")
    #     telemetry = Telemetry(strategy)
    #     telemetry.set_filter("game")
    #     data = telemetry.get_data()
    #     self.assertIsInstance(data, list)

    def test_games(self):
        telemetry = Telemetry()
        games = telemetry.games()
        game_names = [game.name for game in games]  # Accessing the game name attribute directly
        self.assertIn("iRacing", game_names)

    def test_games_pandas(self):
        telemetry = Telemetry()
        telemetry.set_pandas_adapter()
        games = telemetry.games()
        self.assertIsInstance(games, pd.DataFrame)



import pandas as pd
from typing import Optional

def average_speed(df: pd.DataFrame, speed_column: str = 'SpeedMs') -> Optional[float]:
    """
    Calculate the average speed from a DataFrame.

    Args:
        df (pd.DataFrame): The DataFrame containing the telemetry data.
        speed_column (str): The name of the column containing speed data. Defaults to 'Speed'.

    Returns:
        Optional[float]: The average speed, or None if the speed column is not found or contains no valid data.
    """
    if speed_column not in df.columns:
        print(f"Warning: '{speed_column}' column not found in the DataFrame.")
        return None

    average = df[speed_column].mean()
    if pd.isna(average):
        print(f"Warning: No valid data found in the '{speed_column}' column.")
        return None

    return average



import plotly.graph_objects as go
import hashlib


def telemetry_for_fig(segment, track_length=None):
    if segment.start > segment.end:
        # add track_length to all distances that are less than start
        df = segment.telemetry.copy()
        if track_length is None:
            track_length = df["DistanceRoundTrack"].max()
            print(f"track_length: {track_length}")
        df["DistanceRoundTrack"] = df["DistanceRoundTrack"].apply(
            lambda x: x + track_length if x < segment.start else x
        )
        return df
    return segment.telemetry


def features_for_fig(segment, track_length, features):
    if segment.start > segment.end:
        features = features.copy()
        for key in ["start", "end", "max_start", "max_end"]:
            value = features[key]
            if value < segment.start:
                # print(f"adding track_length to {key} {value} -> {value + track_length}")
                features[key] = value + track_length
    return features


def get_color_from_string(s):
    """Generate a color based on the hash of the input string, with defaults for Brake and Throttle."""
    if s == "Brake":
        return "red"
    elif s == "Throttle":
        return "green"
    else:
        hash_object = hashlib.md5(s.encode())
        hash_hex = hash_object.hexdigest()
        r, g, b = int(hash_hex[:2], 16), int(hash_hex[2:4], 16), int(hash_hex[4:6], 16)
        return f'rgb({r},{g},{b})'

def lap_fig(df, mode=None, columns=["Throttle", "Brake"], fig=None,
            full_range=True, title=None, show_x_axis=False,
            show_legend=True):
    # Add title with TrackCode and CarModel if not provided
    if title is None:
        title = ""
        track_code = df["TrackCode"].iloc[0]
        car_model = df["CarModel"].iloc[0]
        if track_code:
            title += f"Track: {track_code}"
        if car_model:
            title += f" - Car: {car_model}" if title else f"Car: {car_model}"

    layout_base = {
        # 'height': 100,  # Increased height to accommodate title
        'xaxis': {
            'showgrid': True,
            'zeroline': False,
            'gridcolor': '#E2E2E2',
            'side': 'top',
            'fixedrange': False,
            'visible': show_x_axis,  # Make x-axis visibility optional
            # 'showticklabels': True,  # Show tick labels only when x-axis is visible
        },
        'yaxis': {
            'showline': False,
            'gridcolor': '#E2E2E2',
            'fixedrange': True,
            'title': title
        },
        'margin': {
            'l': 50,
            'r': 0,
            'b': 10,
            't': 5,  # Increased top margin for title
            'pad': 4
        },
        'paper_bgcolor': '#ffffff',
        'plot_bgcolor': '#ffffff'
    }

    fig = fig or go.Figure(layout=layout_base)
    if fig:
        fig.update_layout(layout_base)

    for column in columns:
        color = get_color_from_string(column)
        fig.add_scatter(
            x=df["DistanceRoundTrack"],
            y=df[column],
            marker=dict(size=1),
            mode=mode,
            name=column,
            line=dict(color=color),
            showlegend=show_legend,
        )

    # Set the range of the x-axis and the distance between tick marks
    # set start to the nearest 100 meters
    if not full_range:
        start = df["DistanceRoundTrack"].min()
        start = start - (start % 100)
        # end = df["DistanceRoundTrack"].max()
        # if end - start < 400:
        #     end = start + 400
        end = start + 1000
        x_range = [start, end]
        fig.update_xaxes(range=x_range, dtick=100)

    return fig

def plot_histogram(df, column_name):
    """
    Create a histogram using Plotly.

    :param df: DataFrame containing the data
    :param column_name: Name of the column to plot
    :return: Plotly Figure object
    """
    value_counts = df[column_name].value_counts().sort_values(ascending=False)

    fig = go.Figure(data=[go.Bar(
        x=value_counts.index,
        y=value_counts.values,
        text=value_counts.values,
        textposition='auto',
    )])

    fig.update_layout(
        title=f'Histogram of {column_name}',
        xaxis_title=column_name,
        yaxis_title='Count',
        bargap=0.2,
    )

    return fig


def fig_add_shape(fig, color="black", **kwargs):
    default = dict(
        type="rect",
        xref="x",
        yref="y",
        x0=0,
        y0=0,
        x1=0,
        y1=1,
        line=dict(color=color, width=2, dash="dot"),
    )
    args = {**default, **kwargs}
    fig.add_shape(**args)
    return fig


def fig_add_features(fig, features, color="red"):
    fig_add_shape(fig, x0=features["start"], x1=features["end"], color=color)
    fig_add_shape(
        fig,
        x0=features["max_start"],
        y0=features["max_low"],
        x1=features["max_end"],
        y1=features["max_high"],
        color=color,
    )
    fig_add_shape(
        fig,
        type="line",
        x0=features["max_start"],
        y0=features["force"],
        x1=features["max_end"],
        y1=features["force"],
        line=dict(color="yellow", width=2),
    )

def plot_3d_map(df):
    """
    Create a 3D map using Plotly with WorldPosition coordinates.

    :param df: DataFrame containing WorldPosition_x, WorldPosition_y, and WorldPosition_z columns
    :return: Plotly Figure object
    """
    fig = go.Figure(data=[go.Scatter3d(
        x=df['WorldPosition_x'],
        y=df['WorldPosition_y'],
        z=df['WorldPosition_z'],
        mode='lines',
        line=dict(
            color='blue',
            width=2
        )
    )])

    fig.update_layout(
        scene=dict(
            xaxis_title='X Position',
            yaxis_title='Y Position',
            zaxis_title='Z Position',
            aspectmode='data'
        ),
        title='3D Map of World Positions'
    )

    return fig

def plot_2d_map(data, landmarks=None):
    """
    Create a 2D map using Plotly with WorldPosition coordinates and optional landmarks.

    :param data: DataFrame or list of DataFrames containing WorldPosition_x, WorldPosition_y, and DistanceRoundTrack columns
    :param landmarks: DataFrame containing landmark information (optional)
    :return: Plotly Figure object
    """
    fig = go.Figure()

    if isinstance(data, list):
        colors = ['blue', 'green', 'red', 'purple', 'orange', 'cyan', 'magenta', 'yellow']
        for i, df in enumerate(data[1:]):
            fig.add_trace(go.Scatter(
                x=df['WorldPosition_x'],
                y=df['WorldPosition_y'],
                mode='lines',
                line=dict(
                    color=colors[i % len(colors)],
                    width=2
                ),
                name=f'Track {i+1}'
            ))
        df = data[0]
    else:
        df = data

    # if landmarks is None or landmarks does not contain any kind == 'segment' rows
    if landmarks is None or landmarks[landmarks['kind'] == 'segment'].empty:
        fig.add_trace(go.Scatter(
            x=df['WorldPosition_x'],
            y=df['WorldPosition_y'],
            mode='lines',
            line=dict(
                color='blue',
                width=2
            ),
            name='Track'
        ))

    if landmarks is not None:
        segment_colors = ['red', 'blue']  # Alternating colors for segments
        color_index = 0

        for _, landmark in landmarks.iterrows():
            if landmark['kind'] == 'turn':
                # Find the closest point to the landmark's start
                closest_point = df.iloc[(df['DistanceRoundTrack'] - landmark['start']).abs().argsort()[:1]]

                fig.add_trace(go.Scatter(
                    x=[closest_point['WorldPosition_x'].values[0]],
                    y=[closest_point['WorldPosition_y'].values[0]],
                    mode='markers',
                    marker=dict(
                        size=10,
                        color='red',
                        symbol='star'
                    ),
                    name=f"{landmark['name']}"
                ))
            elif landmark['kind'] == 'segment':
                # Color points between start and end of the segment
                segment_df = df[(df['DistanceRoundTrack'] > landmark['start']) &
                                (df['DistanceRoundTrack'] <= landmark['end'])]

                fig.add_trace(go.Scatter(
                    x=segment_df['WorldPosition_x'],
                    y=segment_df['WorldPosition_y'],
                    mode='lines',
                    marker=dict(
                        size=2,
                        color=segment_colors[color_index % 2],
                    ),
                    name=f"{landmark['name']}"
                ))

                color_index += 1


    # Calculate the range for x and y axes
    x_min, x_max = df['WorldPosition_x'].min(), df['WorldPosition_x'].max()
    y_min, y_max = df['WorldPosition_y'].min(), df['WorldPosition_y'].max()

    # Add a small margin (e.g., 5%) to the range
    margin = 0.05
    x_range = [x_min - margin * (x_max - x_min), x_max + margin * (x_max - x_min)]
    y_range = [y_min - margin * (y_max - y_min), y_max + margin * (y_max - y_min)]

    fig.update_layout(
        xaxis_title='X Position',
        yaxis_title='Y Position',
        title='2D Map of World Positions with Landmarks',
        xaxis=dict(
            scaleanchor="y",
            scaleratio=1,
            range=x_range,
        ),
        yaxis=dict(
            scaleanchor="x",
            scaleratio=1,
            range=y_range,
        )
    )

    return fig



import os
from typing import Optional, Any, List, Dict
from gql import Client, gql
from gql.dsl import DSLSchema, DSLQuery, dsl_gql
from gql.transport.requests import RequestsHTTPTransport
from .retrieval_strategy import RetrievalStrategy

class GraphQLRetrievalStrategy(RetrievalStrategy):
    def __init__(self, endpoint: str = "") -> None:
        if not endpoint:
            endpoint = "http://telemetry.b4mad.racing:30050/graphql"
        self.endpoint: str = endpoint
        path_to_this_file: str = os.path.dirname(os.path.abspath(__file__))
        path_to_schema: str = os.path.join(path_to_this_file, 'schema.graphql')
        with open(path_to_schema) as f:
            schema_str: str = f.read()
        transport: RequestsHTTPTransport = RequestsHTTPTransport(url=self.endpoint, verify=True, retries=3)
        self.client: Client = Client(transport=transport, schema=schema_str)
        self.ds: DSLSchema = DSLSchema(self.client.schema)

    def retrieve_data(self, filters: Dict[str, Any]) -> List[Dict[str, Any]]:
        if not filters:
            raise Exception("Filters cannot be empty")

        result: List[Dict[str, Any]] = response.get("allTelemetryGames", {}).get("nodes", [])
        return result

    def _query(self, query: Any) -> Dict[str, Any]:
        # query.args(first=10)
        query = dsl_gql(DSLQuery(query))
        return self.client.execute(query)

    def games(self) -> List[Dict[str, Any]]:
        query = self.query_all_games()
        result = self._query(query)
        result = result.get("allTelemetryGames", {}).get("nodes", [])
        return result

    def query_all_games(self) -> Any:
        ds = self.ds
        query = ds.Query.allTelemetryGames
        # query.select(ds.TelemetryGamesConnection.totalCount)
        query.select(ds.TelemetryGamesConnection.nodes.select(ds.TelemetryGame.name))
        return query

    def sessions(self, group_by: Optional[str] = None) -> List[Dict[str, Any]]:
        ds = self.ds
        query = ds.Query.allTelemetrySessions
        if group_by:
            # query.select(ds.TelemetrySessionsConnection.nodes.select(
            #     ds.TelemetrySession.telemetryDriverByDriverId.select(
            #         ds.TelemetryDriver.name
            #     )
            # ))
            # also count the sessions
            query.select(ds.TelemetrySessionsConnection.nodes.select(
                ds.TelemetrySession.count
            ))
        else:
            query.select(ds.TelemetrySessionsConnection.nodes.select(
                ds.TelemetrySession.sessionId,
                ds.TelemetrySession.start,
                ds.TelemetrySession.end,
                # ds.TelemetrySession.telemetryDriverByDriverId.select(
                #     ds.TelemetryDriver.name
                # ),
            ))
        result = self._query(query)
        result = result.get("allTelemetrySessions", {}).get("nodes", [])
        return result



import csv
from .adapter import Adapter

class CSVAdapter(Adapter):
    def convert(self, data):
        # Implement conversion to CSV here
        output = []
        for row in data:
            output.append(','.join(map(str, row)))
        return '\n'.join(output)



from typing import Optional, Dict, Callable, List
import math

class Streaming:
    def __init__(self, average_speed: bool = False, coasting_time: bool = False, raceline_yaw: bool = False, ground_speed: bool = False, braking_point: bool = False, **kwargs):
        self.total_speed: float = 0.0
        self.count: int = 0
        self.features: Dict[str, Callable] = {}
        self.computed_features: Dict[str, float] = {}

        if average_speed:
            self.configure_feature("average_speed", lambda telemetry: self.average_speed(telemetry.get("SpeedMs", 0)))
        if coasting_time:
            self.configure_feature("coasting_time", self.coasting_time)
        if raceline_yaw:
            self.configure_feature("raceline_yaw", self.raceline_yaw)
        if ground_speed:
            self.configure_feature("ground_speed", self.ground_speed)
        if braking_point:
            self.configure_feature("braking_point", self.braking_point)

        self.last_lap_time: float = 0.0
        self.last_x: float = 0.0
        self.last_y: float = 0.0
        self.total_coasting_time: float = 0.0
        self.brake_pressed: bool = False
        self.last_brake_pressure: float = 0.0
        self.braking_point_found: bool = False
        self.brake_pressure_threshold: float = 0.1
        self.rate_of_change_threshold: float = 0.05

    def configure_feature(self, name: str, feature_func: Callable):
        """
        Configure a new feature to be computed.

        Args:
            name (str): The name of the feature.
            feature_func (Callable): The function to compute the feature.
        """
        self.features[name] = feature_func
        self.computed_features[name] = 0.0

    def notify(self, telemetry: Dict):
        """
        Process new telemetry data and compute configured features.

        Args:
            telemetry (Dict): The incoming telemetry data.
        """
        self.elapsed_time = self._calculate_elapsed_time(telemetry.get("CurrentLapTime", 0.0))
        self.dx, self.dy = self._calculate_position_delta(
            telemetry.get("WorldPosition_x", 0.0),
            telemetry.get("WorldPosition_y", 0.0)
        )

        for feature_name, feature_func in self.features.items():
            result = feature_func(telemetry)
            self.computed_features[feature_name] = result

    def get_features(self) -> Dict[str, float]:
        """
        Get the computed features.

        Returns:
            Dict[str, List[float]]: A dictionary of feature names and their computed values.
        """
        return self.computed_features

    def _calculate_elapsed_time(self, current_lap_time: float) -> float:
        """Calculate elapsed time since last update."""
        if self.last_lap_time == 0:
            self.last_lap_time = current_lap_time
            return 0
        elapsed_time = current_lap_time - self.last_lap_time
        self.last_lap_time = current_lap_time
        return elapsed_time

    def _calculate_position_delta(self, current_x: float, current_y: float) -> tuple[float, float]:
        """Calculate the change in position since last update."""
        if self.last_x == 0 and self.last_y == 0:
            self.last_x, self.last_y = current_x, current_y
            return 0.0, 0.0
        dx = current_x - self.last_x
        dy = current_y - self.last_y
        self.last_x, self.last_y = current_x, current_y
        return dx, dy

    def average_speed(self, current_speed: float) -> float:
        """
        Calculate the running average speed.

        Args:
            current_speed (float): The current speed value.

        Returns:
            Optional[float]: The updated average speed, or None if no data has been processed.
        """
        self.total_speed += current_speed
        self.count += 1

        if self.count == 0:
            return 0.0

        return self.total_speed / self.count

    def coasting_time(self, telemetry: Dict) -> float:
        """
        Calculate the time spent coasting (no Throttle or Brake applied).

        Args:
            telemetry (Dict): The incoming telemetry data.

        Returns:
            float: The total time spent coasting in seconds.
        """
        if telemetry.get("Throttle", 0) == 0 and telemetry.get("Brake", 0) == 0:
            self.total_coasting_time += self.elapsed_time
        return self.total_coasting_time

    def raceline_yaw(self, telemetry: Dict) -> float:
        """
        Calculate the yaw based on the current and previous x and y coordinates.

        Args:
            telemetry (Dict): The incoming telemetry data.

        Returns:
            float: The calculated yaw angle between -180 and 180 degrees.
        """

        dx, dy = self.dx, self.dy

        if dx == 0 and dy == 0:
            return 0.0

        yaw = math.degrees(math.atan2(dy, dx))

        yaw = (yaw - 90) % 360
        if yaw > 180:
            yaw -= 360

        return yaw

    def ground_speed(self, telemetry: Dict) -> float:
        """
        Calculate the ground speed based on x and y coordinates traveled between ticks.

        Args:
            telemetry (Dict): The incoming telemetry data.

        Returns:
            float: The calculated ground speed in meters per second.
        """

        if self.elapsed_time == 0:
            return 0.0

        dx, dy = self.dx, self.dy

        distance = math.sqrt(dx**2 + dy**2)
        speed = distance / self.elapsed_time

        return speed

    def braking_point(self, telemetry: Dict) -> float:
        """
        Determine the braking point based on brake pressure and its rate of change.

        Args:
            telemetry (Dict): The incoming telemetry data.

        Returns:
            float: The DistanceRoundTrack when the braking point is detected, or -1 if not yet detected.
        """
        if self.braking_point_found:
            return self.computed_features["braking_point"]

        current_brake_pressure = telemetry.get("Brake", 0)

        time_difference = self.elapsed_time

        if time_difference > 0:
            rate_of_change = (current_brake_pressure - self.last_brake_pressure) / time_difference

            if (current_brake_pressure > self.brake_pressure_threshold and
                rate_of_change > self.rate_of_change_threshold):
                self.braking_point_found = True
                return telemetry.get("DistanceRoundTrack", -1)

        self.last_brake_pressure = current_brake_pressure
        return -1




from setuptools import setup, find_packages

setup(
    name='racing-telemetry',
    version='0.1.4',
    packages=find_packages(include=['racing_telemetry', 'racing_telemetry.*']),
    install_requires=[
        'gql',
        'requests',
        'influxdb-client',
        'pandas',
        'requests-toolbelt',
        'psycopg2-binary',
        'sqlalchemy',
    ],
    author='Marcel Hild',
    author_email='hild@b4mad.net',
    description='A library for telemetry data analysis',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/b4mad/racing-telemetry',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
)



import os
import sys
from typing import Optional
from sqlalchemy import create_engine, func, text
from sqlalchemy.orm import sessionmaker
from sqlalchemy.ext.automap import automap_base
from .retrieval_strategy import RetrievalStrategy

class PostgresRetrievalStrategy(RetrievalStrategy):
    def __init__(self):
        dbname = os.getenv('DB_NAME', 'postgres')
        user = os.getenv('DB_USER', 'postgres')
        password = os.getenv('DB_PASSWORD', 'postgres')
        host = os.getenv('DB_HOST', 'localhost')
        port = os.getenv('DB_PORT', '5432')

        self.engine = create_engine(f'postgresql://{user}:{password}@{host}:{port}/{dbname}')
        self.db_session = sessionmaker(bind=self.engine)

        Base = automap_base()
        Base.prepare(self.engine, reflect=True)

        self.Game = Base.classes.telemetry_game
        self.Session = Base.classes.telemetry_session
        self.Driver = Base.classes.telemetry_driver
        self.Track = Base.classes.telemetry_track
        self.Landmark = Base.classes.telemetry_landmark
        self.Lap = Base.classes.telemetry_lap
        self.Car = Base.classes.telemetry_car

    def retrieve_data(self, query):
        with self.db_session() as session:
            result = session.execute(text(query)).fetchall()
        return result

    def games(self):
        with self.db_session() as session:
            return session.query(self.Game).all()

    def drivers(self):
        with self.db_session() as session:
            return session.query(self.Driver).all()

    def sessions(self, limit: Optional[int] = 10, group_by=None, game_name=None, track_name=None, driver_name=None):
        with self.db_session() as session:
            query = session.query(self.Session)

            if game_name:
                query = query.join(self.Game).filter(self.Game.name == game_name)
            if track_name:
                query = query.join(self.Lap).join(self.Track).filter(self.Track.name == track_name)
            if driver_name:
                query = query.join(self.Driver).filter(self.Driver.name == driver_name)

            if not group_by:
                # s = query.statement
                # print(s.compile(compile_kwargs={"literal_binds": True}))
                return query.limit(limit).all()

            query = query.with_entities(func.count(self.Session.id).label('count'))

            if group_by == 'game':
                query = query.add_columns(self.Game.name)
                query = query.join(self.Game, self.Session.game_id == self.Game.id)
                group_column = self.Game.name
            elif group_by == 'driver':
                query = query.add_columns(self.Driver.name)
                query = query.join(self.Driver, self.Session.driver_id == self.Driver.id)
                group_column = self.Driver.name
            elif group_by == 'track':
                query = query.add_columns(self.Track.name)
                query = query.join(self.Lap).join(self.Track)
                group_column = self.Track.name
            else:
                group_column = getattr(self.Session, group_by)
                query = query.add_columns(group_column)

            return query.group_by(group_column).limit(limit).all()

    def tracks(self, game_name=None, track_name=None):
        with self.db_session() as session:
            query = session.query(self.Track)

            if game_name:
                query = query.join(self.Game).filter(self.Game.name == game_name)

            if track_name:
                query = query.filter(self.Track.name == track_name)

            return query.all()

    def laps(self, limit: Optional[int] = 10, session_id=None, track_name=None, driver_name=None):
        with self.db_session() as session:
            query = session.query(self.Lap)

            if session_id:
                query = query.filter(self.Lap.session_id == session_id)
            if track_name:
                query = query.join(self.Track).filter(self.Track.name == track_name)
            if driver_name:
                query = query.join(self.Session).join(self.Driver).filter(self.Driver.name == driver_name)

            query = query.order_by(self.Lap.start.desc())

            return query.limit(limit).all()

    def landmarks(self, game_name=None, track_name=None, kind=None):
        with self.db_session() as session:
            query = session.query(self.Landmark)

            if game_name or track_name:
                query = query.join(self.Track)

                if game_name:
                    query = query.join(self.Game).filter(self.Game.name == game_name)

                if track_name:
                    query = query.filter(self.Track.name == track_name)

            if kind:
                query = query.filter(self.Landmark.kind == kind)

            return query.all()

    def cars(self, game=None):
        with self.db_session() as session:
            query = session.query(self.Car)

            if game:
                query = query.join(self.Game).filter(self.Game.name == game)

            return query.all()



import datetime

from dirtyfields import DirtyFieldsMixin
from django.db import models
from django_prometheus.models import ExportModelOperationsMixin
from model_utils.models import TimeStampedModel
from picklefield.fields import PickledObjectField


class Driver(ExportModelOperationsMixin("driver"), TimeStampedModel):
    class Meta:
        ordering = [
            "name",
        ]

    name = models.CharField(max_length=200, unique=True)

    def __str__(self):
        return self.name


class Game(TimeStampedModel):
    class Meta:
        ordering = [
            "name",
        ]

    name = models.CharField(max_length=200, unique=True)

    def __str__(self):
        return self.name


class Track(TimeStampedModel):
    class Meta:
        ordering = [
            "name",
        ]

    name = models.CharField(max_length=200)
    length = models.IntegerField(default=0)

    game = models.ForeignKey(Game, on_delete=models.CASCADE, related_name="tracks")

    def __str__(self):
        return self.name


class Car(TimeStampedModel):
    class Meta:
        ordering = [
            "name",
        ]

    name = models.CharField(max_length=200)

    game = models.ForeignKey(Game, on_delete=models.CASCADE, related_name="cars")
    car_class = models.ForeignKey("CarClass", on_delete=models.CASCADE, related_name="cars", null=True)

    def __str__(self):
        return self.name


class CarClass(TimeStampedModel):
    class Meta:
        ordering = [
            "name",
        ]

    name = models.CharField(max_length=200)

    game = models.ForeignKey(Game, on_delete=models.CASCADE, related_name="car_classes")

    def __str__(self):
        return self.name


class SessionType(TimeStampedModel):
    type = models.CharField(max_length=200, unique=True)

    def __str__(self):
        return self.type


class Session(ExportModelOperationsMixin("session"), DirtyFieldsMixin, TimeStampedModel):
    session_id = models.CharField(max_length=200)
    start = models.DateTimeField(default=datetime.datetime.now)
    end = models.DateTimeField(default=datetime.datetime.now)

    driver = models.ForeignKey(Driver, on_delete=models.CASCADE, related_name="sessions")
    session_type = models.ForeignKey(SessionType, on_delete=models.CASCADE, related_name="sessions")
    game = models.ForeignKey(Game, on_delete=models.CASCADE, related_name="sessions")

    class Meta:
        unique_together = (
            "driver",
            "session_id",
            "session_type",
            "game",
        )

    def __str__(self):
        return self.session_id


class FastLap(ExportModelOperationsMixin("fastlap"), TimeStampedModel):
    game = models.ForeignKey(Game, on_delete=models.CASCADE, related_name="fast_laps")
    car = models.ForeignKey(Car, on_delete=models.CASCADE, related_name="fast_laps")
    track = models.ForeignKey(Track, on_delete=models.CASCADE, related_name="fast_laps")
    driver = models.ForeignKey(Driver, on_delete=models.CASCADE, related_name="fast_laps", null=True)
    # add binary field to hold arbitrary data
    data = PickledObjectField(null=True)

    class Meta:
        ordering = ["game", "car", "track"]

    def __str__(self):
        return f"{self.id}: {self.game} {self.car} {self.track}"


class Lap(ExportModelOperationsMixin("lap"), DirtyFieldsMixin, TimeStampedModel):
    number = models.IntegerField()
    start = models.DateTimeField(default=datetime.datetime.now)
    end = models.DateTimeField(default=datetime.datetime.now)
    time = models.FloatField(default=0)
    length = models.IntegerField(default=0)
    valid = models.BooleanField(default=False)

    session = models.ForeignKey(Session, on_delete=models.CASCADE, related_name="laps")
    track = models.ForeignKey(Track, on_delete=models.CASCADE, related_name="laps")
    car = models.ForeignKey(Car, on_delete=models.CASCADE, related_name="laps")
    fast_lap = models.ForeignKey(FastLap, on_delete=models.CASCADE, related_name="laps", null=True)

    class Meta:
        ordering = [
            "number",
        ]
        unique_together = ("session", "start")

    def __str__(self):
        return (
            f"{self.number}: {self.start.strftime('%H:%M:%S')} - {self.end.strftime('%H:%M:%S')} "
            + f"{self.time}s {self.length}m valid: {self.valid}"
        )

    def time_human(self):
        minutes = int(self.time // 60)
        seconds = round(self.time % 60, 2)
        # milliseconds = int((coach_lap_time % 1) * 1000)
        time_string = ""
        if minutes > 1:
            time_string += f"{minutes} minutes "
        elif minutes == 1:
            time_string += f"{minutes} minute "

        time_string += f"{seconds:.2f} seconds "

        return time_string


class FastLapSegment(TimeStampedModel):
    turn = models.CharField(max_length=200)
    start = models.IntegerField(default=0)
    end = models.IntegerField(default=0)
    brake = models.IntegerField(default=0)
    turn_in = models.IntegerField(default=0)
    force = models.IntegerField(default=0)
    gear = models.IntegerField(default=0)
    # stop is the time the brake force starts to decrease
    stop = models.IntegerField(default=0)
    accelerate = models.IntegerField(default=0)
    speed = models.IntegerField(default=0)
    mark = models.CharField(max_length=256, default="")

    fast_lap = models.ForeignKey(FastLap, on_delete=models.CASCADE, related_name="fast_lap_segments")

    def __str__(self):
        repr = f"{self.turn}: {self.start} - {self.end} brake: {self.brake} "
        repr += f"turn_in: {self.turn_in} force: {self.force} gear: {self.gear} stop: {self.stop} "
        repr += f"acc: {self.accelerate} speed: {self.speed} mark: {self.mark}"
        return repr


class Coach(ExportModelOperationsMixin("coach"), TimeStampedModel):
    driver = models.OneToOneField(
        Driver,
        on_delete=models.CASCADE,
        primary_key=True,
    )

    error = models.TextField(default="")
    status = models.TextField(default="")
    enabled = models.BooleanField(default=False)

    fast_lap = models.ForeignKey(FastLap, on_delete=models.SET_NULL, related_name="coaches", null=True)

    MODE_DEFAULT = "default"
    MODE_TRACK_GUIDE = "track_guide"
    MODE_TRACK_GUIDE_APP = "track_guide_app"
    MODE_DEBUG_APP = "debug_app"
    MODE_DEBUG = "debug"
    MODE_ONLY_BRAKE = "only_brake"
    MODE_ONLY_BRAKE_DEBUG = "only_brake_debug"
    MODE_COPILOTS = "copilots"
    MODE_CHOICES = [
        (MODE_DEFAULT, "Default"),
        (MODE_COPILOTS, "Copilots"),
        (MODE_TRACK_GUIDE, "Track Guide"),
        (MODE_TRACK_GUIDE_APP, "Track Guide Application"),
        (MODE_DEBUG_APP, "Debug Application"),
        (MODE_DEBUG, "Debug"),
        (MODE_ONLY_BRAKE, "Only Brakepoints"),
        (MODE_ONLY_BRAKE_DEBUG, "Only Brakepoints (Debug))"),
    ]
    mode = models.CharField(max_length=64, default=MODE_DEFAULT, choices=MODE_CHOICES)

    def __str__(self):
        return self.driver.name


class Landmark(TimeStampedModel):
    name = models.CharField(max_length=200)
    start = models.IntegerField(null=True)
    end = models.IntegerField(null=True)
    is_overtaking_spot = models.BooleanField(null=True)
    from_cc = models.BooleanField(default=False)

    KIND_MISC = "misc"
    KIND_SEGMENT = "segment"
    KIND_TURN = "turn"
    KIND_CHOICES = [
        (KIND_MISC, "Misc"),
        (KIND_SEGMENT, "Segment"),
        (KIND_TURN, "Turn"),
    ]

    kind = models.CharField(max_length=64, default=KIND_MISC, choices=KIND_CHOICES)

    track = models.ForeignKey(Track, on_delete=models.CASCADE, related_name="landmarks")

    def __str__(self):
        return self.name


class TrackGuide(TimeStampedModel):
    name = models.CharField(max_length=200)
    description = models.TextField(default="")
    car = models.ForeignKey(Car, on_delete=models.CASCADE)
    track = models.ForeignKey(Track, on_delete=models.CASCADE)

    def car_game(self):
        return self.car.game

    def __str__(self):
        return f"{self.name}"


class TrackGuideNote(TimeStampedModel):
    track_guide = models.ForeignKey(TrackGuide, on_delete=models.CASCADE, related_name="notes")

    landmark = models.ForeignKey(Landmark, on_delete=models.CASCADE, related_name="notes", null=True)
    segment = models.IntegerField(default=0)
    finish_at = models.TextField(null=True)
    at = models.CharField(max_length=64, default="")

    priority = models.IntegerField(default=0)
    ref_id = models.CharField(max_length=64, default="")
    ref_eval = models.CharField(max_length=64, default="")

    sort_key = models.CharField(max_length=64, default="")
    mode = models.CharField(max_length=64, default="")

    message = models.TextField(default="")
    eval = models.TextField(default="")
    notes = models.TextField(default="")
    score = models.CharField(max_length=64, default="")

    def __str__(self):
        repr = f"{self.segment or self.landmark} - at {self.at or self.finish_at}: {self.message}"
        return repr


class SoundClip(TimeStampedModel):
    subtitle = models.TextField()
    voice = models.CharField(max_length=200)
    model = models.CharField(max_length=200)
    audio_file = models.FileField(upload_to="sound_clips/")

    def __str__(self):
        return f"{self.voice} - {self.subtitle[:50]}"
